import inspect
import traceback

from beepy.utils.js_py import js, log, to_js


class const_attribute(property):
    def __set__(self, instance, value):
        if self.__get__(instance) is None:
            super().__set__(instance, value)
        else:
            raise AttributeError


def _debugger(error=None):
    if isinstance(error, Exception):
        log.warn(traceback.format_exc())
        error_frame = tuple(traceback.walk_tb(error.__traceback__))[-1][0]
    else:
        log.warn(''.join(traceback.format_stack()[:-1]))
        log.warn(error)
        error_frame = None

    frame = inspect.currentframe().f_back
    js._locals = to_js(frame.f_locals)
    js._locals._frame = frame
    js._locals.error_frame = error_frame
    js._DEBUGGER(error)


__all__ = ['const_attribute', '_debugger']



from typing import Any

from boltons.typeutils import make_sentinel

from beepy.attrs import attr, html_attr, state
from beepy.children import Children
from beepy.framework import Tag
from beepy.utils import __CONFIG__, js
from beepy.utils.common import AnyOfType, get_random_name

AUTO_ID = make_sentinel(var_name='AUTO_ID')


class html_tag(Tag, _root=True, content_tag=None):
    contenteditable = html_attr(type=bool)
    id = html_attr(type=str)
    class_ = html_attr(type=str)

    def __set_ref__(self, parent, ref):
        super().__set_ref__(parent, ref)
        if type(self).id is html_tag.id and (self.id is AUTO_ID or (__CONFIG__['inputs_auto_id'] and self.id is None)):
            # TODO: replace 5 and 2 with some log value
            self.id = f'{ref.name or get_random_name(5)}-{get_random_name(2)}'


def by__input_id(input_tag):
    return input_tag.id.rsplit('-', 1)[0]


def by__ref(tag):
    if ref := tag._ref:
        return ref.name


class div(html_tag, name='div', content_tag=None):
    pass


class hr(html_tag, name='hr'):
    pass


class img(html_tag, name='img'):
    src = html_attr(type=str)


class a(html_tag, name='a'):
    href = attr(type=str)
    target = attr(enum={'_blank', '_self', '_parent', '_top', AnyOfType(str)}, type=str)


class p(html_tag, name='p'):
    pass


class b(html_tag, name='b'):
    pass


class i(html_tag, name='i'):
    pass


class ul(html_tag, name='ul'):
    pass


class li(html_tag, name='li'):
    pass


class span(html_tag, name='span'):
    pass


class table(html_tag, name='table'):
    pass


class thead(html_tag, name='thead'):
    pass


class tbody(html_tag, name='tbody'):
    pass


class tr(html_tag, name='tr'):
    pass


class th(html_tag, name='th'):
    pass


class td(html_tag, name='td'):
    pass


class label(html_tag, name='label'):
    for_ = attr(type=str)

    def __set_ref__(self, parent, ref):
        super().__set_ref__(parent, ref)
        if parent and isinstance(self.for_, Tag):
            self.for_ = self.for_._ref.__get__(parent).id


class form(html_tag, name='form'):
    pass


class h1(html_tag, name='h1'):
    pass


class h2(html_tag, name='h2'):
    pass


class h3(html_tag, name='h3'):
    pass


class h4(html_tag, name='h4'):
    pass


class h5(html_tag, name='h5'):
    pass


class h6(html_tag, name='h6'):
    pass


class input_(html_tag, name='input'):
    type = attr(
        'text',
        enum=(
            {'button', 'number', 'checkbox', 'color', 'date', 'datetime-local', 'email', 'file', 'hidden', 'image'}
            | {'password', 'radio', 'range', 'reset', 'search', 'submit', 'tel', 'text', 'time', 'url', 'week', 'month'}
        ),
    )
    hidden = attr(default=False)
    value = html_attr('', model='input')
    placeholder = html_attr(None, type=str)

    def clear(self):
        self.value = ''


class change(input_):
    value = html_attr('', model='change')


class textarea(html_tag, name='textarea'):
    value = html_attr('', model='change')
    data_gramm = html_attr(default=True, type=str)

    def clear(self):
        self.value = ''


class header(html_tag, name='header'):
    pass


class main(html_tag, name='main'):
    pass


class footer(html_tag, name='footer'):
    pass


class nav(html_tag, name='nav'):
    pass


class button(html_tag, name='button'):
    type = attr('button', enum={'submit', 'reset', 'button'}, type=str)


class option(html_tag, name='option'):
    value = html_attr()
    label = html_attr('')
    defaultSelected = html_attr(default=False)
    selected = html_attr(default=False)


class select(html_tag, name='select'):
    value = html_attr(model='change')

    children = [
        options := Children(),
    ]

    def select(self, value):
        self.value = value

    @classmethod
    def with_items(cls, items: dict[str, Any], **kwargs):
        return cls(options=[option(label=label_, value=value) for value, label_ in items.items()], **kwargs)


class StandaloneTag(html_tag, _root=True):
    def __init__(self, *args, **kwargs):
        kwargs['_load_children'] = True
        super().__init__(*args, **kwargs)

    def __mount__(self, element, parent: Tag, index=None):
        # too many copy-paste?
        self.parent = parent
        self.mount_parent = element
        self.pre_mount()

    def clone(self, parent=None):  # noqa: ARG002 - arguments for overriding
        return self

    def as_child(self, parent: Tag | None, *, exists_ok=False, inline_def=False):  # noqa: ARG002 - args for overriding
        return super().as_child(parent, exists_ok=True, inline_def=inline_def)


class Head(StandaloneTag, name='head', mount=js.document.head):
    title = state()

    def render(self):
        if self.title:
            js.document.title = self.title


Head = Head()


class Body(StandaloneTag, name='body', mount=js.document.body):
    style = html_attr()


Body = Body()


# TODO: add all HTML tags


__all__ = ['html_tag', 'div', 'a', 'p', 'b', 'i', 'ul', 'li', 'span', 'form', 'h1', 'h2', 'h3', 'h4', 'h5', 'h6']
__all__ += ['input_', 'textarea', 'header', 'main', 'footer', 'nav', 'button', 'option', 'select', 'Head']



from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Iterable
from functools import wraps
from typing import TYPE_CHECKING

from beepy.trackable import TrackableList
from beepy.utils import __CONFIG__, js
from beepy.utils.common import escape_html

if TYPE_CHECKING:
    from beepy.children import ChildrenRef
    from beepy.framework import Tag


class AttrValue:
    """
    Extend this class to be able to use it as value in beepy.attr and children
    """

    __slots__ = ('value',)

    def __init__(self, value):
        self.value = value

    def __repr__(self):
        return f'AttrValue[{self.value}]'

    def __str__(self):
        return repr(self)

    @abstractmethod
    def __view_value__(self):
        """This method will be called on render, must return serializable value"""


class safe_html(str):
    def __html__(self) -> str:
        return self


def safe_html_content(function):
    @wraps(function)
    def content_wrapper(*args, **kwargs):
        return safe_html(function(*args, **kwargs))

    return content_wrapper


class Renderer:
    __slots__ = ()

    def _render(self, value: str | Iterable[str]) -> str:
        if isinstance(value, safe_html):
            return value.__html__()

        if isinstance(value, str):
            return escape_html(value, whitespace=__CONFIG__['html_replace_whitespaces'])

        if isinstance(value, Iterable):
            return ''.join(self._render(child) for child in value)

        return str(value)

    @abstractmethod
    def __render__(self, *a, **kw):
        # leave here any arguments to haven't problems when override this method
        pass


class Mounter:
    # TODO: add mount_element, mount_parent, etc?
    __slots__ = ()

    @abstractmethod
    def __mount__(self, element: js.HTMLElement, parent: Mounter, index: int | None = None):
        pass

    @abstractmethod
    def __unmount__(self, element: js.HTMLElement, parent: Mounter):
        pass


class WebBase(Renderer, Mounter, ABC):
    __slots__ = ()


class Children(WebBase, TrackableList):
    # TODO: extend Children from Context too?
    __slots__ = ('parent', 'parent_index', 'ref', 'mounted')

    parent: Tag | None
    parent_index: int
    ref: ChildrenRef | None
    mounted: bool

    def __init__(self, iterable=()):
        super().__init__(iterable)
        self.parent = None
        self.parent_index = 0
        self.ref = None
        self.mounted = False

    def as_child(self, parent: Tag | None, *, exists_ok=False):
        from beepy.children import ChildrenRef

        if self.ref:
            if exists_ok:
                return self.ref
            else:
                raise TypeError(f'{self} already is child')
        ref = ChildrenRef(self)
        self.__set_parent__(parent, 0, ref)
        return ref

    def __set_parent__(self, parent: Tag | None, index: int, ref: ChildrenRef):
        self.parent = parent
        self.parent_index = index
        self.ref = ref
        if ref.inline_def:
            setattr(type(parent), ref.name, self)

    def onchange_notify(self):
        if not self.parent or not self.parent.parent:
            return

        for trigger in self.onchange_triggers:
            trigger(self.parent)

    def _notify_add_one(self, key: int, child: Tag):
        if not self.mounted and not self.parent:
            return

        child.link_parent_attrs(self.parent)
        child.__mount__(self.parent._children_element, self.parent, key + self.parent_index)
        if self.parent._mount_finished_:
            child.__render__()

    def _notify_remove_one(self, _key: int, child: Tag):
        if not self.mounted and not self.parent:
            return

        child.__unmount__(self.parent._children_element, self.parent)

    def __render__(self):
        for child in self:
            child.__render__()

    def __mount__(self, element, parent: Tag, index=None):
        self.mounted = True

        if index is not None:
            index += self.parent_index
        for child in self:
            child.link_parent_attrs(parent)
            child.__mount__(element, parent, index)

    def __unmount__(self, element, parent):
        self.mounted = False

        for child in self:
            child.__unmount__(element, parent)


AttrType = None | str | int | bool | Iterable['AttrType'] | dict[str, 'AttrType'] | AttrValue  # Also: Tag | attr
ContentType = str | Iterable['ContentType'] | Renderer

__all__ = ['AttrType', 'ContentType', 'AttrValue', 'Renderer', 'Mounter', 'Children', 'safe_html', 'safe_html_content']



from __future__ import annotations

from abc import ABCMeta
from typing import TYPE_CHECKING, TypeVar

from boltons.typeutils import make_sentinel

import beepy
from beepy.attrs import attr
from beepy.utils import js
from beepy.utils.common import get_random_name, log10_ceil, to_kebab_case
from beepy.utils.dev import const_attribute
from beepy.utils.js_py import Interval, create_once_callable

if TYPE_CHECKING:
    from beepy.types import AttrType

_base_obj_dir = (*dir(object()), '__abstractmethods__')


Self = TypeVar('Self', bound='Context')

_CONTEXT_INITIALIZED = False

OVERWRITE = make_sentinel('_OVERWRITE', var_name='OVERWRITE')
SUPER = make_sentinel('_SUPER', var_name='SUPER')
CONTENT = make_sentinel('_CONTENT', var_name='CONTENT')
_SPECIAL_CHILD_STRINGS = (OVERWRITE, SUPER, CONTENT)


class _MetaContext(ABCMeta):
    _to_load_before_top_render: dict[Context | None, list] = {None: []}
    _wait_onload_interval: dict[Context, Interval] = {}
    _context_classes = []
    __clean_class_attribute_names = ()
    _current_render = {None: []}  # to prevent ValueError, for now
    _contexts: list[Context]

    def __new__(mcs, _name: str, bases: tuple, namespace: dict, **kwargs):
        initialized = _CONTEXT_INITIALIZED  # if class Context is already defined

        # used for checking inheritance: attributes, methods, etc.
        # for example: extending classes Tag and WithRouter must produce correct state 'router'
        base_cls: type[Context] | type = type.__new__(mcs, _name, bases, {})

        namespace = namespace.copy()
        namespace.setdefault('__slots__', ())
        static_attrs = {}
        attrs_defaults = {}

        if initialized and hasattr(beepy, 'children'):
            _children = beepy.children

            mcs._update_namespace_with_extra_attributes(namespace)

            for attribute_name, child in mcs._clean_namespace(namespace):
                if isinstance(child, attr):
                    static_attrs[attribute_name] = child

            for attribute_name, child in mcs._clean_cls_iter(base_cls):
                new_attr = namespace.get(attribute_name)
                if attribute_name not in static_attrs and isinstance(child, attr):
                    static_attrs[attribute_name] = child

                if new_attr and (
                    (isinstance(child, attr) and not isinstance(new_attr, attr))
                    or (isinstance(child, _children.ChildrenRef) and not isinstance(new_attr, _children.ChildrenRef))
                ):
                    attrs_defaults[attribute_name] = namespace.pop(attribute_name)

        is_root = kwargs.get('_root')
        ctx_name = '' if is_root else kwargs.get('name')
        namespace['__ROOT__'] = is_root

        if ctx_name or (initialized and not hasattr(base_cls, '_context_name_')):
            namespace['_context_name_'] = ctx_name or to_kebab_case(_name)

        cls: type[Context] | type = super().__new__(mcs, _name, bases, namespace)

        if initialized:
            cls._static_attrs = attr.order_dict_by_priority(cls._static_attrs.copy() | static_attrs)
            cls._attrs_defaults = cls._attrs_defaults.copy() | attrs_defaults
        else:
            cls._static_attrs = {}
            cls._attrs_defaults = {}

        cls._contexts = []

        mcs._context_classes.append(cls)

        return cls

    @classmethod
    def _update_namespace_with_extra_attributes(mcs, namespace):
        if not hasattr(beepy, 'children'):
            return

        for _attribute_name, child in mcs._clean_namespace(namespace):
            if isinstance(child, beepy.children.ChildRef | Context) and (
                extra := getattr((child if isinstance(child, Context) else child.child), '__extra_attributes__', None)
            ):
                extra = {key: value for key, value in extra.items() if key not in namespace}
                namespace.update(extra)

    @classmethod
    def _top_mount(mcs, element, root, parent):
        root.style = 'visibility: hidden'
        mcs._current_render[parent] = []
        element.__mount__(root, parent)

    @classmethod
    def _top_render(mcs, element):
        mcs._wait_onload_interval[element._root_parent] = Interval(mcs.wait_onload, (element,), period=0.2)

    @classmethod
    def _top_render_real(mcs, element):
        element._root_parent.mount_element.style = ''
        js.beepy.stopLoading()
        element.__render__()

    @classmethod
    def _clean_namespace(mcs, namespace):
        base_obj_dir = _base_obj_dir + mcs.__clean_class_attribute_names
        for key, value in tuple(namespace.items()):
            if key not in base_obj_dir:
                yield key, value

    @classmethod
    def _clean_cls_iter(mcs, cls):
        base_obj_dir = _base_obj_dir + mcs.__clean_class_attribute_names
        for key in dir(cls):
            if key not in base_obj_dir:
                yield key, getattr(cls, key)

    @classmethod
    def create_onload(mcs):
        @create_once_callable
        def onload(*_, **__):
            mcs._to_load_before_top_render[None].remove(onload)

        mcs._to_load_before_top_render[None].append(onload)
        return onload

    @classmethod
    def wait_onload(mcs, element):
        if not mcs._to_load_before_top_render[None] and not mcs._to_load_before_top_render.get(element._root_parent):
            if interval := mcs._wait_onload_interval.get(element._root_parent):
                interval.clear()
                mcs._wait_onload_interval.pop(element._root_parent)
            mcs._top_render_real(element)


class Context(metaclass=_MetaContext, _root=True):
    __slots__ = ('_id_', '_args', '_kwargs', 'attrs')

    __ROOT__ = False

    _id_: str
    _args: tuple[AttrType, ...]
    _kwargs: dict[str, AttrType]

    _static_attrs: dict[str, attr]
    _attrs_defaults: dict[str, AttrType]
    attrs: dict[str, attr]
    _context_name_: str

    def __new__(cls, *args, **kwargs):
        parent = kwargs.pop('__parent__', None)
        self = super().__new__(cls)
        self.attrs = self._static_attrs.copy()

        # define some attributes here, not in __init__, because they are used for __hash__ method
        self._id_ = get_random_name(log10_ceil(len(self._contexts) * len(self.__class__._context_classes)))
        self._args = args
        self._kwargs = kwargs

        if not _CONTEXT_INITIALIZED:
            return self

        for name, attribute in self.attrs.items():
            attribute._link_cmpt(name, self, force=False)
            if parent and name in kwargs:
                attribute.__set_first__(self, kwargs[name], parent)

        if parent:
            self.link_parent_attrs(parent)

        return self

    def link_parent_attrs(self, parent):
        p_args, p_kwargs = parent.args_kwargs
        p_data = parent._attrs_defaults | p_kwargs

        for name, attr_to_move_on in parent.attrs.items():
            if attr_to_move_on.move_on:
                attr_to_move_on._link_cmpt(name, self, force_cls_set=True)
                if name in p_data:
                    self._attrs_defaults[name] = p_data[name]

    def __init__(self, *args, **kwargs: AttrType):
        self.__class__._contexts.append(self)
        data = self._attrs_defaults | kwargs
        self.init(*args, **data)

    def init(self, *_args, **kwargs):
        for key, attr_ in self.attrs.items():
            if key not in kwargs:
                if attr_.required:
                    raise TypeError(f'Attribute {attr_.name!r} is required')
                continue

            value = kwargs[key]
            setattr(self, key, value)
            attr_.__init_ctx__(self, value)

        self.post_init()

    def post_init(self):
        pass

    @property
    def __view_attrs__(self) -> dict[str, AttrType]:
        return {_attr.name: _attr._get_view_value(self) for _attr in self.attrs.values() if _attr._view}

    @property
    def __attrs__(self) -> dict[str, AttrType]:
        return {
            _attr.name: _attr._get_view_value(self)
            for _attr in self.attrs.values()
            if _attr._view and not _attr._set_on_render
        }

    @property
    def __states__(self) -> dict[str, AttrType]:
        return {_attr.name: _attr.__get__(self) for _attr in self.attrs.values()}

    @const_attribute
    def args_kwargs(self):
        if not hasattr(self, '_args'):
            return None
        # TODO: make self._kwargs as frozendict
        return self._args, self._kwargs

    def __hash__(self):
        # TODO: make force immutable this attributes
        return hash((self._context_name_, self._id_))

    def __notify__(self, attr_name: str, attribute: attr, value: AttrType):
        pass

    def clone(self, parent=None) -> Self:
        args, kwargs = self.args_kwargs
        if parent is not None:
            kwargs = kwargs | {'__parent__': parent}
        return type(self)(*args, **kwargs)


_CONTEXT_INITIALIZED = True


__all__ = ['OVERWRITE', 'SUPER', 'CONTENT', '_SPECIAL_CHILD_STRINGS', '_MetaContext', 'Context']



from __future__ import annotations

from beepy.attrs import attr, state
from beepy.listeners import on
from beepy.style import Style
from beepy.tags import Body, div, hr, li, ul
from beepy.utils import js


class MenuDivider(hr):
    default_style = Style(
        border_bottom='1px solid #eee',
        margin='10px 0',
    )


class MenuItem(li):
    default_style = Style(
        styles={
            'padding': '0.5em 1em',
            'cursor': 'pointer',
            'display': 'flex',
            'align-items': 'center',
            '&:hover': {
                'background-color': '#2d2d2d',
                'border-left': '4px solid #333',
            },
        }
    )


class ContextMenu(ul):
    visible = attr(default=False)

    pos_x = state(0)
    pos_y = state(0)

    dynamic_styles = Style(
        left='{pos_x}px',
        top='{pos_y}px',
    )

    default_style = Style(
        styles={
            'visibility': 'hidden',
            'background-color': '#353535',
            'font-size': '24px',
            'border-radius': '2px',
            'padding': '5px 0 5px 0',
            'width': '180px',
            'height': 'auto',
            'margin': 0,
            'z-index': 50,
            # use absolute positioning
            'position': 'absolute',
            'list-style': 'none',
            'box-shadow': '0 0 20px 0 #222',
            'opacity': 0,
            'transition': '0.2s',
            '&[visible]': {
                'opacity': 1,
                'visibility': 'visible',
            },
        }
    )

    def show(self):
        self.visible = True

    def hide(self):
        self.visible = False

    def set_pos(self, x=0, y=0):
        el = self.mount_element

        if x + el.offsetWidth > js.innerWidth:
            x -= el.offsetWidth

        if y + el.offsetHeight > js.innerHeight:
            y -= el.offsetHeight

        self.pos_x = x
        self.pos_y = y
        self.show()

    def __mount__(self, element, parent, index=None):
        super().__mount__(Body.mount_element, Body)

    def __unmount__(self, element, parent, *, _unsafe=False):
        return super().__unmount__(element, parent, _unsafe=True)


class ContextMenuHandler(div):
    menu: ContextMenu

    default_style = Style(
        position='absolute',
        inset=0,
        width='100%',
        height='100%',
        display='flex',
        justifyContent='center',
        alignItems='center',
    )

    @on
    def click(self):
        self.menu.hide()

    @on('contextmenu.prevent')
    def show_menu(self, event):
        self.menu.set_pos(event.pageX, event.pageY)


__all__ = ['MenuDivider', 'MenuItem', 'ContextMenu', 'ContextMenuHandler']



#!/usr/bin/env python3
import asyncio
import functools
import http.server
import os
import re
import socketserver
import sys
import time
from pathlib import Path
from threading import Thread

import dotenv
from watchdog.events import FileSystemEventHandler
from watchdog.observers import Observer
from websockets.exceptions import ConnectionClosedOK
from websockets.server import serve

dotenv.load_dotenv()


class MonitorFolder(FileSystemEventHandler):
    def __init__(self, server):
        self.server = server

    def on_any_event(self, event):
        if not (
            event.event_type in ('opened', 'closed')
            or event.is_directory
            or event.src_path.endswith(('~', '.tmp'))
            or re.search(r'/(__pycache__|.idea|dist|build)/', event.src_path)
        ):
            self.server.handle_file_event(event)


class DevServer:
    def __init__(self):
        self.websockets = []
        self.root_path = None
        self.observer = None
        self.developer_mode = 'DEVELOPMENT' in os.environ

    async def ws_send(self, message):
        await asyncio.sleep(1)  # Hack for Django autoreload

        for ws in self.websockets[:]:
            try:
                await ws.send(message)
            except ConnectionClosedOK:
                self.websockets.remove(ws)

        if not self.websockets:
            print('[BeePy] No clients connected! Please, restart your page to connect to the dev server')

    def handle_file_event(self, event):
        path: str = event.src_path.removeprefix(self.root_path).removeprefix('/')

        print(f'[BeePy] Found file change: {path}')
        if self.developer_mode:
            if path.endswith('.py'):
                os.system('hatch build')
            elif path.endswith('.js'):
                os.system(f'cd {self.root_path}/web; npm run build; cd -')  # rebuild dist
            asyncio.run(self.ws_send('__'))
        else:
            asyncio.run(self.ws_send(path))

    def _watcher_start(self):
        event_handler = MonitorFolder(self)
        observer = Observer()
        observer.schedule(event_handler, path=self.root_path, recursive=True)
        print(f'[BeePy] Monitoring started for {self.root_path}')
        observer.start()
        self.observer = observer
        try:
            while True:
                time.sleep(1)
        except KeyboardInterrupt:
            observer.stop()
            observer.join()

    async def _ws_echo(self, websocket):
        self.websockets.append(websocket)

        async for message in websocket:
            print(f'{websocket=} {message=}')  # We really don't receive messages

    async def _ws_main(self):
        print('[BeePy] WebSockets started')
        async with serve(self._ws_echo, 'localhost', 8998):
            await asyncio.Future()  # run forever

    def _ws_start(self):
        asyncio.run(self._ws_main())

    def _simple_http_start(self, port=8888):
        with socketserver.TCPServer(
            ('', port), functools.partial(http.server.SimpleHTTPRequestHandler, directory=self.root_path)
        ) as httpd:
            print(f'[BeePy] Serving at port {port}\nOpen server: http://localhost:{port}')
            httpd.serve_forever()

    def start(self, start_http=None, root_path=None, *, forever=True):
        if self.observer is not None:
            print('[BeePy] Server is already started')
            return

        if root_path is None:
            root_path = sys.argv[1] if len(sys.argv) > 1 else Path.cwd()

        self.root_path = root_path

        Thread(target=self._ws_start, daemon=True).start()
        Thread(target=self._watcher_start, daemon=True).start()
        if start_http is not None:
            if forever:
                start_http()
            else:
                Thread(target=start_http, daemon=True).start()


dev_server = DevServer()
start_simple_dev_server = functools.partial(dev_server.start, start_http=dev_server._simple_http_start, forever=True)


if __name__ == '__main__':
    start_simple_dev_server()



import dataclasses
import json
from datetime import datetime
from http import HTTPStatus
from http.client import HTTPException

from beepy.utils.internal import __CONFIG__
from beepy.utils.js_py import IN_BROWSER


class UpgradedJSONEncoder(json.JSONEncoder):  # TODO: consider on rewriting json.JSONEncoder
    def default(self, o):
        if dataclasses.is_dataclass(o):
            return dataclasses.asdict(o)
        if isinstance(o, datetime):
            return o.strftime(__CONFIG__['default_datetime_format'])
        return super().default(o)


if IN_BROWSER:
    from pyodide.http import pyfetch

    async def request(url, method='GET', body=None, headers=None, **opts):
        if body is not None:
            body = json.dumps(body, cls=UpgradedJSONEncoder)

        if headers is None:
            headers = {}

        headers.update(
            {
                'mode': 'no-cors',
                'Content-Type': 'application/json',
                'Access-Control-Allow-Headers': '*',
                'Access-Control-Allow-Origin': '*',
            }
        )

        response = await pyfetch(__CONFIG__['api_url'] + url, method=method, body=body, headers=headers, **opts)

        try:
            response.raise_for_status()
        except OSError as err:
            raise HTTPException(err) from err

        return (await response.json()) if method in ('GET', 'PUT', 'POST') else (await response.text())

else:
    import requests

    async def request(url, method='GET', body=None, headers=None, **_opts):
        if body is not None:
            body = json.dumps(body, cls=UpgradedJSONEncoder)

        if headers is None:
            headers = {}

        # TODO: check opts argument compatibility

        response = requests.request(method, __CONFIG__['api_url'] + url, data=body, headers=headers)

        if int(response.status_code) >= HTTPStatus.BAD_REQUEST:
            raise HTTPException(response.status_code)

        return response.json() if method in ('GET', 'PUT', 'POST') else response.text


__all__ = ['request', 'UpgradedJSONEncoder']



import re
from collections.abc import Iterable
from typing import Any

from beepy.framework import __CONFIG__, Tag, attr, state
from beepy.tags import Head
from beepy.types import AttrValue, safe_html, safe_html_content
from beepy.utils import js
from beepy.utils.common import MISSING, get_random_name, log10_ceil, safe_issubclass, to_kebab_case


def dict_of_properties_to_css(properties):
    for prop, value in properties.items():
        if not prop:
            yield value
            continue
        if not isinstance(prop, Iterable) or isinstance(prop, str):
            prop = (prop,)  # noqa: PLW2901
        if value == '':
            value = '""'  # noqa: PLW2901
        result = ''
        for _property in prop:
            _property = to_kebab_case(_property)
            result += '    ' + _property
            if isinstance(value, tuple | list):  # handle raw css
                result += ' ' + (_property + ' ').join(x for x in value if x)
            else:
                result += f': {value};'
        yield result


def dict_of_parents_to_css(children, parent, braces):
    for child, value in children.items():
        if not child:
            yield f' {braces[0]} {child.strip()} {braces[1]}'
        if isinstance(child, tuple | set):
            child = ','.join(child)  # noqa: PLW2901
        child = re.sub('&', parent, child) if '&' in child else parent + ' ' + child  # noqa: PLW2901
        for inner in dict_to_css_iter(value, child, braces):
            if child.startswith('@') and not parent:
                yield f' {braces[0]} {inner.strip()} {braces[1]}'
            else:
                yield inner


def dict_to_css_iter(selectors: dict, parent: str = '', braces=('{', '}')):
    """
    parses nested dict to css

    dict(
        div=dict(
            color='red',
            backgroundColor='blue',
            p={
                'font-size': '20px',
                '&.class': dict(
                    font_size='30px',
                ),
                '': '''
                    // some raw css
                '''
            }
        )
    )

    to

    div {
        color: red;
        background-color: blue;
    }
    div p {
        font-size: 20px;
                    // some raw css
    }
    div p.class {
        font-size: 30px;
    }


    """

    # TODO: implement all scss or create wheel of libsass, add micropip.import and refactor styles

    children = {}
    properties = {}

    for prop, value in selectors.items():
        if isinstance(value, dict):
            children[prop] = value
        else:
            properties[prop] = value

    if properties:
        yield parent + ' ' + braces[0]
        yield from dict_of_properties_to_css(properties)
        yield braces[1]

    yield from dict_of_parents_to_css(children, parent, braces)


def dict_to_css(*args, separator='\n', **kwargs):
    return separator.join(dict_to_css_iter(*args, **kwargs))


def get_reference(tag: Tag):
    return f'[style-id="{tag.style_id}"]'


class Var:
    __slots__ = ('name',)

    def __init__(self, name):
        self.name = name

    def __str__(self):
        return f'var(--{self.name})'


class StyleRef(AttrValue):
    __slots__ = ('vars', '_style')

    def __init__(self, style_id, style):
        super().__init__(style_id)
        self.vars = {}
        self._style = style

    def __repr__(self):
        return self.value

    __view_value__ = __repr__


class Raw:
    def __bool__(self):
        return False


class Style(Tag, name='style', content_tag=None, raw_html=True, force_ref=True):
    __slots__ = ('styles', '_main_style', 'real_parent')

    __extra_attributes__ = {
        'style_id': attr(type=StyleRef),
    }

    _global = {
        'styles_count': 1,
    }

    options = state(type=dict)
    real_parent: Tag | None

    @classmethod
    def from_css(cls, _file):
        # TODO: implement import from css/scss/pycss
        raise NotImplementedError

    def __init__(self, styles=None, options=None, get_vars=None, **styles_dict):
        super().__init__()
        if styles and not styles_dict:
            styles_dict = styles
        if options is None:
            options = {}
        self.styles = styles_dict
        self._content = ''
        self._main_style = False
        self.real_parent = None
        self.options = {
            'global': False,
            'render_states': True,
            'render_children': True,
            'get_vars_callback': get_vars,
        } | options

    def __mount__(self, element, parent, index=None):
        self.real_parent = parent
        if __CONFIG__['style_head']:
            super().__mount__(Head.mount_element, Head)
        else:
            super().__mount__(element, parent, index)

    def __unmount__(self, element, parent, *, _unsafe=False):
        return super().__unmount__(element, parent, _unsafe=True)

    def mount(self):
        styles = {Raw(): '{__VARS__}'} | self.styles
        parent = self.real_parent

        if self.options['global']:  # TODO: add example
            self._content = dict_to_css(styles, parent._tag_name_)
            return

        if parent.style_id is None:  # support multiple style children
            self._global['styles_count'] += 1
            style_id = get_random_name(log10_ceil(self._global['styles_count']) + 1)
            parent.style_id = StyleRef(style_id, self)
            self._main_style = True
        else:
            style_id = parent.style_id

        # TODO: is [style-id=] slow? If so, maybe use classes?
        self._content = dict_to_css(styles, f'{parent._tag_name_}[style-id="{style_id}"]', braces=('{{', '}}'))

    @safe_html_content
    def content(self):
        if self.options['global'] or not self._content:
            return self._content

        parent = self.real_parent
        params: dict[str, Any] = {'__VARS__': ''}

        if self.options['render_states']:
            params.update(parent.__states__)

        if self.options['render_children']:
            params.update(parent.ref_children)

        if get_vars := self.options['get_vars_callback']:
            params.update(get_vars(self=parent, ref=get_reference, **params))

        if self._main_style and (vars := parent.style_id.vars):
            # TODO: add mangling for --var names
            params['__VARS__'] = dict_to_css({safe_html(f'--{name}'): var for name, var in vars.items()}).strip()[1:-1]

        # TODO: use native css '--var: {}' instead of re-render the whole content
        return self._content.strip().format(**params)

    def var(self, name, new_value=MISSING):
        parent = self.real_parent
        if not parent:
            return

        if new_value is MISSING:
            return parent.style_id.vars[name]
        else:
            if new_value is None:
                del parent.style_id.vars[name]
            else:
                parent.style_id.vars[name] = new_value
            if parent._mount_finished_:
                parent.__render__()

    @classmethod
    def import_file(cls, file_path):
        return js.beepy.addElement(
            js.document.head,
            'link',
            href=js.beepy.files.getPathWithCurrentPathAndOrigin(file_path),
            onload=cls.create_onload(),
            type='text/css',
            rel='stylesheet',
        )


def with_style(style_or_tag_cls: Style | type[Tag] | None = None):
    """
    @with_style
    class Button(Tag, name='button'):
        ...

    styled_button = with_style(Style(display='block', **kwargs))(button)
    """

    def wrapper(tag_cls: type[Tag]):
        return type(tag_cls.__name__, (tag_cls,), {'style': style_or_tag_cls or Style()})

    if safe_issubclass(style_or_tag_cls, Tag):
        _tag_cls, style_or_tag_cls = style_or_tag_cls, None
        return wrapper(_tag_cls)

    return wrapper


__all__ = ['dict_to_css_iter', 'dict_to_css', 'Style', 'with_style']



import asyncio
import os
import sys
from importlib import import_module
from typing import TYPE_CHECKING

import dotenv
import micropip

from beepy.utils.js_py import IN_BROWSER, js, to_js

if TYPE_CHECKING:
    from collections.abc import Callable

BEEPY_ROOT_PACKAGE = '__beepy_root__'
_PY_TAG_ATTRIBUTE = '__PYTHON_TAG__'

dotenv.load_dotenv(f'{BEEPY_ROOT_PACKAGE}/.env' if IN_BROWSER else '.env')

# TODO: make it with dataclass
__CONFIG__ = {
    'debug': 'DEBUG' in os.environ,
    'development': 'DEVELOPMENT' in os.environ,
    'style_head': True,
    'api_url': '/',
    'default_datetime_format': '%Y-%m-%dT%H:%M:%S.%f%Z',
    'inputs_auto_id': True,
    'html_replace_whitespaces': True,
    'requirements': [],
}


def merge_configs():
    __CONFIG__.update(js.beepy.config.to_py())
    js.beepy.config = to_js(__CONFIG__)


if IN_BROWSER:  # TODO: check support for non-browser runs
    merge_configs()


async def reload_requirements():
    get_requirements: list | Callable = __CONFIG__['requirements']
    if not callable(get_requirements):  # static requirements, must be already loaded
        return

    await asyncio.gather(*[micropip.install(requirement) for requirement in get_requirements()])


def lazy_import(module_path):
    if not (
        (module := sys.modules.get(module_path))
        and (spec := getattr(module, '__spec__', None))
        and getattr(spec, '_initializing', False) is False
    ):
        module = import_module(module_path)

    return module


def import_string(dotted_path):
    module_path, class_name = dotted_path.rsplit('.', 1)
    return getattr(lazy_import(module_path), class_name)


def lazy_import_cls(cls):
    if isinstance(cls, str):
        return import_string(cls)
    return cls


def _init_js():
    from beepy import __version__

    js.console.log(f'%cBeePy version: {__version__}', 'color: lightgreen; font-size: 35px')
    merge_configs()


class _BeePyGlobals(dict):
    def __getitem__(self, key):
        result = super().__getitem__(key)
        for handler in __beepy_global_handlers__:
            result = handler(self, key, result)
        return result


def _default_global_handlers(locals_dict, key, result):
    if key.startswith('_p__'):
        return result(locals_dict)
    return result


__beepy_global_handlers__ = [_default_global_handlers]


__all__ = [
    '_PY_TAG_ATTRIBUTE',
    '__CONFIG__',
    'merge_configs',
    'lazy_import',
    'import_string',
    'lazy_import_cls',
    'reload_requirements',
    '__beepy_global_handlers__',
]



from __future__ import annotations

import traceback
from collections.abc import Callable, Iterable
from functools import cache
from types import MethodType

from beepy.attrs import attr, set_html_attribute, state
from beepy.children import ChildRef, Children, ContentWrapper, CustomWrapper, StringWrapper, TagRef
from beepy.components import Component, _MetaComponent
from beepy.context import _SPECIAL_CHILD_STRINGS, CONTENT, OVERWRITE, SUPER
from beepy.types import AttrType, ContentType, Mounter, Renderer
from beepy.utils import __CONFIG__, js, log
from beepy.utils.common import NONE_TYPE, get_random_name, to_kebab_case
from beepy.utils.dev import _debugger
from beepy.utils.internal import _PY_TAG_ATTRIBUTE

__version__ = '0.8.6'  # For internal development set to 0.0a0
__CONFIG__['version'] = __version__


_TAG_INITIALIZED = False


class _MetaTag(_MetaComponent):
    _tag_classes: list[type[Tag]] = []
    __clean_class_attribute_names = ('_content_tag', '_static_children_tag')

    def __new__(mcs, _name: str, bases: tuple, namespace: dict, **kwargs):  # noqa: PLR0912, PLR0915
        namespace = namespace.copy()
        namespace.setdefault('__slots__', ())
        if '_content' in namespace:
            namespace['_static_content'] = namespace.pop('_content')

        # used for checking inheritance: attributes, methods, etc.
        # for example: extending classes Tag and WithRouter must produce correct state 'router'
        mock_cls: type[Tag] | type = type.__new__(mcs, _name, bases, {})

        initialized = _TAG_INITIALIZED  # As base classes is also declared here, we must be sure base class exists

        is_root = kwargs.get('_root')
        tag_name = '' if is_root else kwargs.get('name')
        namespace['__ROOT__'] = is_root

        if tag_name or (initialized and not hasattr(mock_cls, '_tag_name_')):
            namespace['_tag_name_'] = to_kebab_case(tag_name or _name)

        if 'raw_html' in kwargs:
            namespace['_raw_html'] = kwargs['raw_html']

        if 'force_ref' in kwargs:
            namespace['_force_ref'] = kwargs['force_ref']

        if 'mount' in kwargs:
            namespace['mount_element'] = kwargs['mount']

        if 'content_tag' in kwargs:
            content_tag = kwargs['content_tag']
            if isinstance(content_tag, str):
                content_tag = empty_tag(content_tag)()
            elif not isinstance(content_tag, Tag | NONE_TYPE):
                raise TypeError('content_tag argument must be either str, Tag or None')
            namespace['_content_tag'] = content_tag

        if 'children_tag' in kwargs:
            children_tag = kwargs['children_tag']
            if isinstance(children_tag, str):
                children_tag = empty_tag(children_tag)()
            elif not isinstance(children_tag, Tag):
                raise TypeError('children_tag argument must be either str or Tag')
            namespace['_static_children_tag'] = children_tag

        ref_children = []
        to_remove_children = []

        children_arg = namespace.get('children', [])
        if isinstance(children_arg, property):
            children_arg = []

        if children_arg:
            namespace.pop('children')
            children_arg = list(children_arg)
            if OVERWRITE not in children_arg and SUPER not in children_arg:
                children_arg.insert(0, SUPER)
        elif children_arg not in (None, False, []):
            namespace['_static_children'] = namespace.pop('children')
            children_arg = []

        if initialized:
            mcs._update_namespace_with_extra_attributes(namespace)

            for attribute_name, child in tuple(mcs._clean_namespace(namespace)):
                if callable(child) and hasattr(child, '_attrs_static_'):
                    continue

                if isinstance(child, Component | Children | ChildRef):
                    if child in children_arg:
                        ref_children.append(child)
                        # TODO: make possible inherit without replacement?
                        if (old_child := getattr(mock_cls, attribute_name, None)) and isinstance(old_child, ChildRef):
                            to_remove_children.append(old_child)
                        if isinstance(child, Component | Children):
                            children_arg[children_arg.index(child)] = namespace[attribute_name] = child.as_child(None)

                    if isinstance(child, Component) and child._force_ref:
                        # TODO: add support '_force_ref' for Children too
                        namespace[attribute_name] = _ref = child.as_child(None)
                        _ref.__set_name__(None, attribute_name)
                        children_arg.insert(0, _ref)
                        ref_children.append(_ref)

                    if isinstance(child, ChildRef) and child.child._force_ref:
                        children_arg.insert(0, child)
                        ref_children.append(child)

            for _index, child in enumerate(children_arg):
                if isinstance(child, str) and child not in _SPECIAL_CHILD_STRINGS:
                    children_arg[_index] = StringWrapper(child)

                if isinstance(child, Component) and child not in ref_children:
                    # TODO: replace 5 in get_random_name(5) with some log
                    children_arg[_index] = namespace[f'_{get_random_name(5)}_'] = child.as_child(None)

        cls: type[Tag] | type = super().__new__(mcs, _name, bases, namespace, **kwargs)

        if not hasattr(cls, '_raw_html'):
            cls._raw_html = False

        if initialized:
            cls._static_children = cls._static_children.copy()
            if not hasattr(cls, '_content_tag'):
                cls._content_tag = empty_div()
        else:
            cls._static_children = []

        if not hasattr(cls, '_static_children_tag'):
            cls._static_children_tag = None

        for child in to_remove_children:
            cls._static_children.remove(child)

        if getattr(cls, '_tag_name_', None) or not isinstance(getattr(cls, 'children', None), property):
            if children_arg:
                if CONTENT in children_arg and CONTENT in cls._static_children:
                    cls._static_children.remove(CONTENT)
                if SUPER in children_arg:
                    super_children_index = children_arg.index(SUPER)
                    children_arg[super_children_index : super_children_index + 1] = cls._static_children
                else:
                    children_arg.extend(cls._static_children)
                cls._static_children = children_arg
        else:
            cls._static_children = [CONTENT]

        cls._tags = []

        mcs._tag_classes.append(cls)

        if initialized and 'mount' in kwargs:
            cls._root_parent = None
            setattr(cls.mount_element, _PY_TAG_ATTRIBUTE, cls())

        if cls.__ROOT__:
            cls.__root_declared__()
        else:
            cls.__class_declared__()

        return cls

    @classmethod
    def __META_mount__(cls, self: Tag, args, kwargs, _original_func):
        self.__class__._tags.append(self)

        self._mount_finished_ = False

        result = _original_func(self, *args, **kwargs)

        self._mount_attrs()

        for child in self.children:
            if isinstance(child, CustomWrapper):
                child.__mount__(self.mount_element, self)
            elif isinstance(child, Mounter):
                child.__mount__(self._children_element, self)
            else:
                log.warn(f'Cannot mount: {child}')
        if self._children_tag:
            self.mount_element.insertChild(self._children_element)

        if self.content_child is None:
            _debugger(self)
        else:
            self.content_child._mount_children()

        super().__META_mount__(self, args, kwargs, None, _mount_attrs=False)

        self._mount_finished_ = True

        return result

    @classmethod
    def __META_render__(cls, self: Tag, args, kwargs, _original_func):
        if not self._mount_finished_ or (
            not self.mount_element.parentElement and self._root_parent != self  # dismounted
        ):
            return  # Prevent render before mount finished; Could be useful for setting intervals inside mount method

        return super().__META_render__(self, args, kwargs, _original_func)

    @classmethod
    def __META_init__(cls, self: Tag, args, kwargs, _original_func):  # noqa: PLR0912, PLR0915
        kwargs.pop('__parent__', None)  # called from clone
        if hasattr(getattr(self, 'mount_element', None), _PY_TAG_ATTRIBUTE):
            return

        self._children = self._static_children.copy()
        self._content = self._static_content

        children_argument: Callable | ContentType = kwargs.get('children') or args
        if children_argument and (not isinstance(children_argument, Iterable) or isinstance(children_argument, str)):
            children_argument = [children_argument]
        children_argument = list(children_argument)

        if children_argument:
            is_child_arg_string = False
            is_child_arg_function = False
            is_child_arg_tag = False
            try:
                content_empty = not bool(self.content())
            except Exception:  # noqa: BLE001 - trying to call content function
                content_empty = False
            if content_empty:
                is_child_arg_string = True
                is_child_arg_function = True
                is_child_arg_tag = True
                for index, child_arg in enumerate(children_argument):
                    if not isinstance(child_arg, str):
                        is_child_arg_string = False
                    if (isinstance(child_arg, type) and issubclass(child_arg, Tag)) or not callable(child_arg):
                        is_child_arg_function = False
                    if not isinstance(child_arg, Tag):
                        if isinstance(child_arg, str) and len(children_argument) > 1:
                            children_argument[index] = empty_span(child_arg)
                        else:
                            is_child_arg_tag = False

            if is_child_arg_string and children_argument:
                self._content = children_argument
            elif is_child_arg_function and children_argument and len(children_argument) == 1:
                self._content = MethodType(children_argument[0], self)
            elif is_child_arg_tag:
                self._children = self._children.copy() + [
                    child.clone(self).as_child(self) for child in children_argument
                ]
            else:
                self._children = [*self._children.copy(), *children_argument]

        for key, value in kwargs.items():
            if not isinstance(value, Component | Children) or key in self.attrs:
                continue
            value: Component | Children
            child = value.as_child(self, exists_ok=True)
            child.__set_name__(self, key)
            setattr(type(self), key, child)
            self._children.append(child)

        self._shadow_root = None
        self.mount_parent = None
        self._mount_finished_ = False

        if not hasattr(self, 'mount_element'):
            self.mount_element = js.document.createElement(self._tag_name_)
        if getattr(self.mount_element, _PY_TAG_ATTRIBUTE, None):
            raise ValueError(f'Coping or using as child is not allowed for "{self._tag_name_}"')
        else:
            setattr(self.mount_element, _PY_TAG_ATTRIBUTE, self)
        if self._static_children_tag:
            self._children_tag = self._static_children_tag.clone(self)
            self._children_element = self._children_tag.mount_element
            setattr(self._children_element, _PY_TAG_ATTRIBUTE, self)
        else:
            self._children_tag = None
            self._children_element = self.mount_element

        super().__META_init__(self, args, kwargs, _original_func)

    @classmethod
    def __META_unmount__(cls, self: Tag, args, kwargs, _original_func):
        if self in self.__class__._tags:  # But why?
            self.__class__._tags.remove(self)

        super().__META_unmount__(self, args, kwargs, None, _post_mount=False)

        for child in self.children:
            if isinstance(child, CustomWrapper):
                child.__unmount__(self.mount_element, self)
            elif isinstance(child, Mounter):
                child.__unmount__(self._children_element, self)
        if self._children_tag:
            self.mount_element.safeRemoveChild(self._children_element)

        result = _original_func(self, *args, **kwargs)

        self.post_unmount()

        return result


class Tag(Component, metaclass=_MetaTag, _root=True):
    # TODO: add docstrings

    __slots__ = (
        '_content',
        '_shadow_root',
        'mount_parent',
        '_mount_finished_',
        'mount_element',
        '_children',
        '_children_element',
        '_children_tag',
    )

    _root_parent: Tag = None  # see function mount in the bottom

    _static_content: ContentType = ''

    _content: ContentType
    _shadow_root: js.HTMLElement
    _ref: TagRef | None
    _force_ref: bool = False

    _tag_name_: str
    _tags: list[Tag]
    mount_element: js.HTMLElement
    mount_parent: js.HTMLElement
    _static_children: list[ContentType]
    _children: list[Mounter]
    _children_element: js.HTMLElement
    _static_children_tag: Tag
    _children_tag: Tag
    _mount_finished_: bool

    @classmethod
    def __root_declared__(cls):
        """This method is called, when root Tag is defined"""

    @classmethod
    def __class_declared__(cls):
        """This method is called, when common Tag is defined"""

    def __init__(self, *args, **kwargs: AttrType):
        kwargs.setdefault('_load_children', False)
        super().__init__(*args, **kwargs)

    def __new__(cls, *args, **kwargs):
        if hasattr(getattr(cls, 'mount_element', None), _PY_TAG_ATTRIBUTE):
            self = getattr(cls.mount_element, _PY_TAG_ATTRIBUTE)
        else:
            self = super().__new__(cls, *args, **kwargs)

        return self

    def __repr__(self):
        return f'{type(self).__name__}(<{self._tag_name_}/>, id#{self._id_})'

    def __hash__(self):
        # TODO: make force immutable this attributes
        return hash((self._tag_name_, self._id_))

    def __getitem__(self, key):
        return getattr(self, key)

    def as_child(self, parent: Tag | None, *, exists_ok=False, inline_def=False):
        if self._ref:
            if exists_ok:
                self.__set_ref__(parent, self._ref)
                return self._ref
            else:
                raise TypeError(f'Tag {self._tag_name_} already is child')
        ref = TagRef(self, inline_def=inline_def)
        self.__set_ref__(parent, ref)
        return ref

    def __notify__(self, attr_name: str, attribute: attr, value: AttrType):
        super().__notify__(attr_name, attribute, value)
        self.__render__()

    def __set_ref__(self, parent: Tag | None, ref: TagRef):
        super().__set_ref__(parent, ref)
        if ref.inline_def:
            setattr(type(parent), ref.name, self)

    def __render__(self, attrs: dict[str, AttrType] | None = None):
        self._current_render.append(self)

        if attrs is None:
            attrs = {}

        for name, value in {**self.__attrs__, **attrs}.items():
            # TODO: optimize this - set only changed attributes

            type = _attr.type if (_attr := self.attrs.get(name)) else NONE_TYPE
            set_html_attribute(self.mount_element, name, value, type=type)

        content_index = self.get_content_index()
        if content_index is not None:
            self.children[content_index].__render__()

        for index, child in enumerate(self.children):
            # TODO: optimize this - re-render the only children, that need this
            if isinstance(child, ChildRef):
                child.__render__(self)
            elif isinstance(child, CustomWrapper):
                if content_index is None or index != content_index:
                    child.__render__()
            elif isinstance(child, Renderer):
                child.__render__()
            else:  # string ?
                self._children_element.append(self._render(child))

        if self._current_render[-1] is self:
            self._current_render.pop()

    def content(self) -> ContentType:
        return self._content() if callable(self._content) else self._content

    def init(self, *args, _load_children=True, **kwargs):
        if _load_children:
            self.children = self.children.copy()

        for name, value in self.ref_children.items():
            if name not in kwargs:
                continue

            if isinstance(value, Children):
                getattr(self, name)[:] = kwargs[name]

        super().init(*args, **kwargs)

    def __mount__(self, element, parent: Tag, index=None):
        self.mount_parent = element

        super().__mount__(element, parent, index=index)

        self.mount_parent.insertChild(self.mount_element, index)

    def __unmount__(self, element, parent, *, _unsafe=False):
        if not _unsafe and self.mount_parent is not None and self.mount_parent is not element:
            log.warn(
                'Something went wrong!\n'
                f'Real parent: {self.mount_parent} {self.parent}. Passed parent: {element} {parent} '
                'If you override __mount__, you also should override __unmount__ too.'
            )
            log.warn(''.join(traceback.format_stack()[:-1]))

        (self.mount_parent or element).safeRemoveChild(self.mount_element)

    @property
    def own_children(self) -> list[Mounter]:
        return [child for child in self.children if not isinstance(child, ChildRef | CustomWrapper)]

    @property
    def ref_children(self) -> dict[str, Mounter]:
        return {child.name: child.__get__(self) for child in self.children if isinstance(child, ChildRef)}

    @property
    def content_child(self) -> ContentWrapper:
        for child in self.children:
            if isinstance(child, ContentWrapper) and child.content.__func__.__name__ == 'content':
                return child

    def get_content_index(self):
        for index, child in enumerate(self.children):
            if isinstance(child, ContentWrapper) and child.content.__func__.__name__ == 'content':
                return index

    @property
    def children(self) -> list[Mounter | Renderer | ChildRef | ContentWrapper]:
        return self._children

    @children.setter
    def children(self, children):
        self._children = self._handle_children(children)

    def _handle_children(self, children):
        result = []

        for new_child in children:
            child = new_child

            if isinstance(child, ContentWrapper):
                child = child.content.__func__
                if child.__name__ == 'content':
                    child = CONTENT

            if child == CONTENT:
                child = self.content

            if isinstance(child, attr):

                def child(s, _n=child.name):
                    return getattr(s, _n)

            if isinstance(child, Component) and child in self._children and child._ref is not None:
                # using Component as descriptor/just child; this allows save reference from parent to new copy of child
                child = child._ref._update_child(self, self._children.index(child))
            elif isinstance(child, ChildRef) and child in self._children:
                child._update_child(self, self._children.index(child))
            # TODO: make ContentWrapper descriptor, here call .clone()
            #  make possible to use ContentWrapper separately from function 'content'
            elif callable(child):
                if not isinstance(child, MethodType):
                    child = MethodType(child, self)
                child = ContentWrapper(child, self._content_tag, self._current_render)

            result.append(child)

        return result

    def clone(self, parent=None) -> Tag:
        clone = super().clone(parent=parent)
        clone._children = self.children
        return clone

    @property
    def _current_render(self):
        try:
            return _MetaTag._current_render[self._root_parent]
        except KeyError as e:
            _debugger(e)
            raise ValueError('It looks like element is not mounted correctly, please see the docs') from None


_TAG_INITIALIZED = True


@cache
def empty_tag(name):
    return _MetaTag(name, (Tag,), {}, name=name, content_tag=None)


def inline_tag(name, content_tag=None, **attributes):
    return _MetaTag(name, (Tag,), attributes, name=name, content_tag=content_tag)


empty_div = empty_tag('div')
empty_span = empty_tag('span')


def mount(element: Tag, root_element: str, *, clear=False):
    root = js.document.querySelector(root_element)
    if root is None:
        raise NameError('Mount point not found')

    js.beepy.stopLoading()
    if clear or js.beepy.dev_server.started:
        root.innerHTML = ''
    js.beepy.startLoading(mountPoint=root)

    parent = inline_tag(root.tagName.lower(), _root_parent=state(type=Tag, move_on=True))()
    parent._root_parent = parent
    parent._attrs_defaults['_root_parent'] = parent
    parent.__class__._root_parent.initial_value = parent
    parent.mount_element = root
    element.link_parent_attrs(parent)

    setattr(root, _PY_TAG_ATTRIBUTE, parent)
    _MetaTag._top_mount(element, root, parent)
    _MetaTag._top_render(element)

    if not js.document.title:
        js.document.title = 'BeePy'
        log.warn(f'Document title is not set, use default title: {js.document.title}')


__all__ = ['__version__', '_MetaTag', 'Tag', 'empty_tag', 'mount']



"""
This file is partly mock of JS API, created to make possible run and debug BeePy apps using console
"""

from __future__ import annotations

import time
from collections import defaultdict
from threading import Thread
from typing import TYPE_CHECKING, Self

try:
    from pyodide.ffi import IN_BROWSER, create_once_callable, create_proxy
except ImportError:
    from pyodide import IN_BROWSER, create_once_callable, create_proxy


class HTMLElement:
    __PYTHON_TAG__: Tag

    def __init__(self, tag_name):
        self.attributes = {}
        self.data = []
        self.listeners = defaultdict(list)
        self.tagName = tag_name
        self.shadowRoot = None
        self.clientWidth = self.clientHeight = self.scrollWidth = self.scrollHeight = 1

    def getAttribute(self, name):
        return self.attributes.get(name)

    def setAttribute(self, name, value):
        self.attributes[name] = value

    def removeAttribute(self, name):
        self.attributes.pop(name, None)

    def append(self, string: str):
        self.data.append(string)
        print(self)

    def appendChild(self, el: HTMLElement):
        self.data.append(el)
        print(self)

    def insertChild(self, el: HTMLElement | str, index: int = None):
        if index is None:
            self.data.append(el)
        else:
            self.data.insert(index, el)
        print(self)

    def removeChild(self, child: HTMLElement):
        self.data.remove(child)

    def safeRemoveChild(self, child: HTMLElement):
        if child in self.data:
            self.data.remove(child)

    def addEventListener(self, name, js_proxy):
        self.listeners[name].append(js_proxy)

    def attachShadow(self, **kwargs):
        self.shadowRoot = HTMLElement('-')
        return self.shadowRoot

    @property
    def innerHTML(self):
        return repr(self)

    @innerHTML.setter
    def innerHTML(self, value):
        self.data = [value]
        print(self)

    def __repr__(self):
        attrs = ' '
        for key, value in self.attributes.items():
            attrs += f'{key}="{value}" '
        attrs = attrs.strip(' ')
        if attrs:
            attrs = ' ' + attrs
        if not self.data:
            return f'<{self.tagName}{attrs}/>'
        return f"<{self.tagName}{attrs}>{''.join(map(str, self.data))}</{self.tagName}>"


class Element(HTMLElement):
    pass


class Console:
    def log(self, *a):
        _ = (self,)
        return print(*a)

    warn = error = debug = info = log


class Document:
    el_cls = Element

    head = el_cls('head')
    body = el_cls('body')

    title = 'BeePy'

    def createElement(self, tag_name):
        return self.el_cls(tag_name)

    querySelector = createElement

    def createDocumentFragment(self):
        return self.createElement('-')


class Location:
    pathname = '/e/'  # must be dynamic
    href = f'http://localhost:9000/{pathname}'
    search = ''
    hash = ''


class History:
    args = ()
    length = 0
    state = {}

    def pushState(self, state, unused, href):
        self.state = state
        self.args = (state, unused, href)

    replaceState = pushState  # We really don't care :)


class SearchParams:
    params = {}

    def get(self, name):
        return self.params.get(name)

    def set(self, name, value):
        self.params[name] = value


class URL:
    hash = '#'

    def __init__(self, href):
        self.href = href
        self.searchParams = SearchParams()

    @classmethod
    def new(cls, *a, **kw):
        return cls(*a, **kw)


class Object:
    @classmethod
    def keys(cls, obj):
        return obj.keys()

    @classmethod
    def fromEntries(cls, objs):
        return objs


class Event:
    def __init__(self, name):
        self.type = name


console = Console()
document = Document()
location = Location()
history = History()


async def delay(ms):
    time.sleep(ms)


class LocalStorage(dict):
    def getItem(self, key):
        return self[key]

    def setItem(self, key, value):
        self[key] = value


localStorage = LocalStorage()


def decodeURI(text):
    return text


def decodeURIComponent(text):
    return text


def _DEBUGGER(error=None):
    print(error)


max_id = {'interval': 1, 'timeout': 1}
threads = {'interval': {}, 'timeout': {}}
listeners = defaultdict(lambda: defaultdict(list))
intervals = {}


def _js_func(fn):
    fn.name = fn.__qualname__
    return fn


@_js_func
def addEventListener(element, name, proxy):
    listeners[element][name].append(proxy)


@_js_func
def removeEventListener(element, name, proxy):
    listeners[element][name].remove(proxy)


@_js_func
def setInterval(callback, ms):
    def interval():
        while True:
            time.sleep(ms / 1000)
            callback()

    id = max_id['interval']
    max_id['interval'] += 1

    threads['interval'][id] = t = Thread(target=interval, daemon=True)
    t.start()


@_js_func
def clearInterval(interval_id):
    if interval_id in threads['interval']:
        threads['interval'][interval_id].join()


@_js_func
def setTimeout(callback, ms):
    def timeout():
        time.sleep(ms / 1000)
        callback()

    id = max_id['timeout']
    max_id['timeout'] += 1

    threads['timeout'][id] = t = Thread(target=timeout, daemon=True)
    t.start()


@_js_func
def clearTimeout(timeout_id):
    if timeout_id in threads['timeout']:
        threads['timeout'][timeout_id].join()


if TYPE_CHECKING:
    from beepy.framework import Tag


class BeePyModule:
    config = {}

    def addElement(self, mount_point, element_name, **options):
        return HTMLElement(element_name)

    def startLoading(self, *a, **kw):
        return

    def stopLoading(self):
        return

    def addAsyncListener(self, el, eventName, method, modifiers, **options):
        return

    class files:
        _lastLoadedFile = ''

        def getPathWithCurrentPathAndOrigin(self, path):
            return path

    class dev_server:
        started = True


beepy = BeePyModule()
_locals = {}
window = self = globalThis = Self


#################
#       .       #
#       .       #
# . . . . . . . #
#       .       #
#       .       #
#################


class Destroyable:
    def destroy(self):
        pass


EVENT_LISTENERS = {}


def add_event_listener(elt, event, listener):
    """Wrapper for JavaScript's addEventListener() which automatically manages the lifetime
    of a JsProxy corresponding to the listener param.
    """
    proxy = create_proxy(listener)
    EVENT_LISTENERS[(elt.js_id, event, listener)] = proxy
    elt.addEventListener(event, proxy)


def remove_event_listener(elt, event, listener):
    """Wrapper for JavaScript's removeEventListener() which automatically manages the lifetime
    of a JsProxy corresponding to the listener param.
    """
    proxy = EVENT_LISTENERS.pop((elt.js_id, event, listener))
    elt.removeEventListener(event, proxy)
    proxy.destroy()


TIMEOUTS: dict[int, Destroyable] = {}


def set_timeout(callback, timeout):
    """Wrapper for JavaScript's setTimeout() which automatically manages the lifetime
    of a JsProxy corresponding to the callback param.
    """
    id = -1

    def wrapper():
        callback()
        TIMEOUTS.pop(id, None)

    callable = create_once_callable(wrapper)
    id = setTimeout(callable, timeout)
    TIMEOUTS[id] = callable
    return id


# An object with a no-op destroy method so we can do
#
# TIMEOUTS.pop(id, DUMMY_DESTROYABLE).destroy()
#
# and either it gets a real object and calls the real destroy method or it gets
# the fake which does nothing. This is to handle the case where clear_timeout is
# called after the timeout executes.
DUMMY_DESTROYABLE = Destroyable()


def clear_timeout(id):
    """Wrapper for JavaScript's clearTimeout() which automatically manages the lifetime
    of a JsProxy corresponding to the callback param.
    """
    clearTimeout(id)
    TIMEOUTS.pop(id, DUMMY_DESTROYABLE).destroy()


INTERVAL_CALLBACKS: dict[int, Destroyable] = {}


def set_interval(callback, interval):
    """Wrapper for JavaScript's setInterval() which automatically manages the lifetime
    of a JsProxy corresponding to the callback param.
    """
    proxy = create_proxy(callback)
    id = setInterval(proxy, interval)
    INTERVAL_CALLBACKS[id] = proxy
    return id


def clear_interval(id):
    """Wrapper for JavaScript's clearInterval() which automatically manages the lifetime
    of a JsProxy corresponding to the callback param.
    """
    clearInterval(id)
    INTERVAL_CALLBACKS.pop(id, DUMMY_DESTROYABLE).destroy()



from beepy.utils.asyncio import ensure_sync, force_sync, sleep
from beepy.utils.internal import __CONFIG__
from beepy.utils.js_py import IN_BROWSER, js, log, to_js

__all__ = ['js', 'log', 'to_js', 'IN_BROWSER', '__CONFIG__', 'ensure_sync', 'force_sync', 'sleep']



from functools import wraps

try:
    import js  # reference to window (globalThis), made by pyodide module
except ModuleNotFoundError:
    from . import js

    js.window = js.self = js.globalThis = js

from pyodide.ffi import IN_BROWSER, create_once_callable, create_proxy
from pyodide.ffi import to_js as pyodide_to_js
from pyodide.ffi.wrappers import (
    add_event_listener,
    clear_interval,
    clear_timeout,
    remove_event_listener,
    set_interval,
    set_timeout,
)

log = js.console


class Interval:
    __slots__ = ('_id',)

    def __init__(self, function, args=None, kwargs=None, period=0):
        @wraps(function)
        def _callback():
            return function(*(args or ()), **(kwargs or {}))

        self._id = set_interval(_callback, period * 1000)

    def clear(self):
        return clear_interval(self._id)


def to_js(obj, dict_converter=js.Object.fromEntries, **kwargs):
    return pyodide_to_js(obj, dict_converter=dict_converter, **kwargs)


def push_url(url, unused='', **url_state):
    js.history.pushState(to_js({'href': url.href} | url_state), unused, url.href)


def replace_url(url, unused='', **url_state):
    js.history.replaceState(to_js({'href': url.href} | url_state), unused, url.href)


__all__ = [
    'js',
    'IN_BROWSER',
    'create_once_callable',
    'create_proxy',
    'set_timeout',
    'clear_timeout',
    'set_interval',
    'clear_interval',
    'add_event_listener',
    'remove_event_listener',
    'Interval',
    'to_js',
    'push_url',
    'replace_url',
]



from __future__ import annotations

from beepy.tags import button


class Action(button, _root=True):
    components: dict[str, type[Action]] = {}

    action_name: str

    @classmethod
    def __class_declared__(cls):
        super().__class_declared__()

        cls.components[cls.action_name] = cls



from __future__ import annotations

import keyword
from collections import defaultdict
from typing import TYPE_CHECKING, Any, get_type_hints

from boltons.iterutils import first
from boltons.typeutils import issubclass, make_sentinel

from beepy.types import AttrType, AttrValue
from beepy.utils import log
from beepy.utils.common import NONE_TYPE, call_handler_with_optional_arguments, to_kebab_case, wraps_with_name

if TYPE_CHECKING:
    import builtins
    from collections.abc import Callable, Sequence
    from typing import TypeVar

    from beepy.components import Component
    from beepy.context import Context

    T = TypeVar('T')

SPECIAL_CONVERT_ATTRIBUTES = {
    'contenteditable': lambda tag, val: (
        convert_boolean_attribute_value(val) if val == tag.mount_element.isContentEditable else val
    )
}


def convert_boolean_attribute_value(value):
    return '' if value else None


_MISSING = make_sentinel(var_name='_MISSING')


def set_html_attribute(el, name: str, value, *, type: builtins.type = NONE_TYPE):
    if value is None:
        el.removeAttribute(name)
    elif name.endswith('_'):
        el.setAttribute(name[:-1], value)
    elif not hasattr(el, name) or issubclass(type, bool):
        el.setAttribute(name, value)
    else:
        setattr(el, name, value)


class attr:
    __slots__ = (
        'name',
        'initial_value',
        'type',
        'const',
        'required',
        'notify',
        'static',
        'move_on',
        'model',
        'model_options',
        '_from_model_cache',
        'fget',
        'fset',
        'fdel',
        'handlers',
        'enum',
        '_cache',
    )

    _view = True
    _set_on_render = False

    name: str | None
    initial_value: T
    type: builtins.type | None
    const: bool
    required: bool
    notify: bool
    static: bool
    move_on: bool
    model: str | None
    model_options: dict[str, Any]
    _from_model_cache: list[tuple[Component, attr, str | None]]
    fget: Callable[[Component], T]
    fset: Callable[[Component, T], None]
    fdel: Callable[[Component], None]
    handlers: dict[str, list[Callable[[Component, T], None]]]
    enum: Sequence

    _cache: dict[Component, AttrType]

    def __init__(  # noqa: PLR0913    # TODO: think about splitting logic
        self,
        default=None,
        *,
        const=False,
        required=False,
        notify=False,
        static=False,
        move_on=False,
        model=None,
        model_options=None,
        fget=None,
        fset=None,
        fdel=None,
        enum=None,
        **kwargs,
    ):
        _type = kwargs.get('type')

        self.name = None
        self.const = const
        assert not const or default is None, f'Const {type(self).__name__} cannot have initial value'
        self.initial_value = default
        self.required = required or const  # const attr must be also required
        self.notify = notify
        self.static = static
        self.move_on = move_on
        self.model = 'change' if model is True else model
        self.model_options = {'attribute': None} | (model_options or {})
        self._from_model_cache = []

        if default is None and _type is None:
            self.type = NONE_TYPE
        else:
            if _type is None:
                _type = default.type if self.model and isinstance(default, attr) else type(default)
            self._set_type(_type)

        # TODO: think on: is this needed?
        # behaviour like @property
        self.fget = fget
        self.fset = fset
        self.fdel = fdel
        if fget:
            self(fget)

        self.handlers = defaultdict(list)
        self.enum = enum

        self._cache = {}

    @property
    def priority(self):
        if self.move_on:
            return 0
        elif self.model:
            return 2
        else:
            return 1

    @classmethod
    def order_dict_by_priority(cls, dict_attrs):
        return dict(sorted(dict_attrs.items(), key=lambda item: item[1].priority))

    def __get__(self, instance, owner=None):
        if instance is None:
            return self
        return (self.fget or self._fget)(instance)

    def _fget(self, instance):
        return self._cache.get(None if self.static else instance, self.initial_value)

    def __call__(self, fget):
        self.fget = fget
        self.name = to_kebab_case(fget.__name__)
        if self.type is NONE_TYPE:
            type_hints = get_type_hints(fget)
            if 'return' in type_hints:
                self._set_type(type_hints['return'])
        return self

    def __set__(self, instance, value, *, _prevent_model=False):
        current_value = self.__get__(instance)
        if self.const and current_value is not None:
            raise AttributeError

        if current_value == value:
            return

        (self.fset or self._fset)(instance, value)

        if instance.parent_defined:
            for handler in self.handlers['change']:
                if _prevent_model and _prevent_model in (True, self) and handler.__name__.startswith('@attr'):
                    continue
                call_handler_with_optional_arguments(handler, instance, {'value': value})

        if self.notify:
            instance.__notify__(self.name, self, value)

    def __set_first__(self, instance, value, parent):
        if self.model and (
            isinstance(value, attr) and value.name is not None and (instance, self, None) not in value._from_model_cache
        ):
            value._from_model_cache.append((instance, self, value.name))
            instance._kwargs.pop(self.name)

    def _fset(self, instance, value):
        if self.fget is not None:
            raise AttributeError

        if self.enum is not None and value not in self.enum:
            raise TypeError(f'Possible values: {self.enum}. Provided value: {value}')

        self._cache[None if self.static else instance] = value

    def setter(self, fset):
        self.fset = fset
        if self.type is NONE_TYPE:
            self._set_type(first(get_type_hints(fset).values()))
        return self

    def __set_name__(self, owner, name):
        if self._view:
            self.name = to_kebab_case(name)
        else:
            self.name = name

    def _set_type(self, _type, *, raise_error=False):
        if hasattr(_type, '__origin__'):  # TODO: add support of strict check of type on change/etc
            _type = _type.__origin__

        if not isinstance(_type, type):
            error = f'Bad type for attribute: {_type!r}, {type(_type)}'
            if raise_error:
                raise TypeError(error)
            else:
                self.type = NONE_TYPE
                return log.error(error)

        self.type = _type

    def _link_cmpt(self, name: str, component: Component, *, force: bool = True, force_cls_set: bool = False):
        component.attrs[name] = self
        if force:
            self.__set_name__(component, name)
        if force_cls_set:
            setattr(type(component), name, self)
        elif not hasattr(component, name) and hasattr(component, 'mount_element'):
            setattr(component, name, self.__get__(component))
        return self

    def _get_type_instance(self):
        try:
            return self.type()
        except Exception as e:  # noqa: BLE001 - trying to call empty constructor
            log.debug(
                f'Got error when trying to empty-args constructor of {self.type}: {e}\n'
                f'Will be better to set {self.name} not to None'
            )

    def _prepare_attribute_for_model(self, instance):
        if attribute := self.model_options['attribute']:
            if callable(attribute):
                return attribute(instance)
            return attribute

    def _set_model_value(self, instance, attr_, component: Component):
        initial_value = self.__get__(instance.parent if instance.parent_defined else component)

        if initial_value is None:
            if self.initial_value is not None:
                initial_value = self.initial_value
            elif self.type and (value_from_type := self._get_type_instance()) is not None:
                initial_value = value_from_type
        if attribute_ := self._prepare_attribute_for_model(instance):
            initial_value = getattr(initial_value, attribute_)
        if initial_value is None:
            initial_value = ''

        attr_.__set__(instance, initial_value, _prevent_model=attribute_)

        return attribute_

    def _handle_model_listeners(self, component: Component):
        to_remove_indexes = []

        for index, (instance, attr_, name) in enumerate(self._from_model_cache):
            if name is None:
                to_remove_indexes.append(index)
                continue

            if instance.parent_defined and instance.parent != component:
                continue

            attribute_ = self._set_model_value(instance, attr_, component)

            _attribute_str = f'({attribute_})' if attribute_ else ''

            @instance.on(attr_.model)
            @wraps_with_name(f'@attr[to:{attr_}->model{_attribute_str}->{self}]')
            def __handler_to_model(parent_instance, event, current_attribute=attribute_):
                value = event.target.value
                if current_attribute:
                    setattr(self.__get__(parent_instance), current_attribute, value)
                else:
                    self.__set__(parent_instance, value, _prevent_model=current_attribute)

            @self.on('change')
            @wraps_with_name(f'@attr[from:{attr_}->model{_attribute_str}->{self}]')
            def __handler_from_model(
                parent_instance, value, current_instance=instance, _attr_=attr_, current_attribute=attribute_
            ):
                instance_ = getattr(parent_instance, current_attribute) if current_attribute else current_instance

                if value is not None and current_attribute:
                    value = getattr(value, current_attribute)

                _attr_.__set__(instance_, value, _prevent_model=current_attribute)

        for index in reversed(to_remove_indexes):
            self._from_model_cache.pop(index)

    def __init_ctx__(self, ctx: Context, value):
        for handler in self.handlers['init']:
            call_handler_with_optional_arguments(handler, ctx, {'value': value})

    def __mount_cmpt__(self, component: Component):
        self._handle_model_listeners(component)

    def __post_mount_cmpt__(self, component: Component):
        for instance, attr_, _ in self._from_model_cache:
            self._set_model_value(instance, attr_, component)

        if self.handlers['mount']:
            value = self.__get__(component)
            for handler in self.handlers['mount']:
                call_handler_with_optional_arguments(handler, component, {'value': value})

    def __delete__(self, instance):
        return (self.fdel or self._fdel)(instance)

    def _fdel(self, instance):
        if self.fget is not None:
            raise AttributeError
        if self.static:
            return

        self._cache.pop(instance, None)

    def deleter(self, fdel):
        self.fdel = fdel
        return self

    def __repr__(self):
        return (
            f'{self.name} = {type(self).__name__}'
            f'(default={self.initial_value!r}, type={self.type}, static={self.static})'
        )

    def __str__(self):
        return f'{self.name}({self.initial_value!r})'

    def on(self, *triggers):
        def wrapper(handler):
            if self.static or self._from_model_cache:  # check if _from_model_cache is required
                if not hasattr(handler, '_attrs_static_'):
                    handler._attrs_static_ = defaultdict(list)
                for trigger in triggers:
                    handler._attrs_static_[trigger].append(self)

            for trigger in triggers:
                if handler in self.handlers[trigger]:
                    raise AttributeError(f"This @on('{trigger}') handler is already set")
            for trigger in triggers:
                self.handlers[trigger].append(handler)
            return handler

        return wrapper

    def _get_view_value(self, instance=None, value=None):
        if instance is not None:
            if value is not None:
                raise ValueError('You cannot provide both instance and value arguments')
            value = self.__get__(instance)

        if check_fn := SPECIAL_CONVERT_ATTRIBUTES.get(self.name):
            return check_fn(instance, value)

        if issubclass(self.type, bool):
            return convert_boolean_attribute_value(value)

        # support for custom types for attr
        if isinstance(self.type, AttrValue):
            return value.__view_value__()

        return value


class state(attr):
    __slots__ = ()

    _view = False


class html_attr(attr):
    __slots__ = ()

    _set_on_render = True

    def __init__(self, *args, name=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.name = name

    def __set_name__(self, owner, name):
        if self.name is None:
            super().__set_name__(owner, name)
        if self.name in keyword.kwlist:
            self.name += '_'

    def _fget(self, instance):
        if instance is None:
            return self

        el = instance.mount_element
        if hasattr(el, self.name):  # noqa: SIM108 - just want to see difference
            value = getattr(el, self.name, None)
        else:
            value = el.getAttribute(self.name)

        return (value == '') if issubclass(self.type, bool) else value

    def _fset(self, instance, value):
        if hasattr(instance, 'mount_element'):
            value = self._get_view_value(value=value)
            set_html_attribute(instance.mount_element, self.name, value, type=self.type)

    def _fdel(self, instance):
        delattr(instance.mount_element, self.name)


class listen_state(state):
    __slots__ = ('provider', 'subscribers')

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.provider = None
        self.subscribers = []

    def __mount_cmpt__(self, component: Component):
        if self.provider is None:
            self.provider = component
        else:
            self.subscribers.append(component)

    def __set__(self, instance, value, *, _prevent_model=False):
        super().__set__(instance, value, _prevent_model=_prevent_model)
        if instance is self.provider:
            for component in self.subscribers:
                component.__notify__(self.name, self, value)


__all__ = ['attr', 'state', 'html_attr']



import base64
import io


def get_plot_img_src(figure):
    buf = io.BytesIO()
    figure.savefig(buf, format='png')
    buf.seek(0)
    data = base64.b64encode(buf.read()).decode('utf-8')
    return f'data:image/png;charset=utf-8;base64,{data}'
